"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var _Database_instances, _Database_timeouts, _Database_loadTimeouts, _Database_createTimeout, _Database_maxAllowedTimeout_get, _Database_sleep;
Object.defineProperty(exports, "__esModule", { value: true });
const tiny_typed_emitter_1 = require("tiny-typed-emitter");
const better_sqlite3_1 = __importDefault(require("better-sqlite3"));
const constants_1 = require("../util/constants");
const Util_1 = require("../util/Util");
const Table_1 = require("./Table");
const DatabaseError_1 = require("./DatabaseError");
const __1 = require("..");
const QuerySelector_1 = require("../builders/QuerySelector");
/**
 * Database wrapper class for sqlite.
 */
class Database extends tiny_typed_emitter_1.TypedEmitter {
    /**
     *
     * @param options The options for this database.
     */
    constructor(options = constants_1.DefaultDatabaseOptions) {
        super();
        _Database_instances.add(this);
        _Database_timeouts.set(this, new Map()
        /**
         * Date at which the database started.
         */
        );
        /**
         * The assigned tables for this database.
         */
        this.tables = new Map();
        this.options = this._validateOptions(Util_1.Util.mergeDefault(options, constants_1.DefaultDatabaseOptions));
    }
    get query() {
        return new QuerySelector_1.QuerySelector(this);
    }
    typedGet(table, options) {
        return this.get(table.toString(), options);
    }
    /**
     * Validates given options.
     * @param options The options to check.
     * @returns The options with their default values.
     */
    _validateOptions(options) {
        return options;
    }
    deleteTimedData(table, key) {
        if (!this.hasTimeoutsTable())
            throw new DatabaseError_1.DatabaseError(`There is not timeouts table instantiated.`);
        const existing = __classPrivateFieldGet(this, _Database_timeouts, "f").get(`${table}_${key}`);
        if (existing) {
            if (existing.timeout)
                clearTimeout(existing.timeout);
            if (existing.sleep)
                clearTimeout(existing.sleep);
            __classPrivateFieldGet(this, _Database_timeouts, "f").delete(`${table}_${key}`);
            this.delete(`__timeouts`, {
                where: [
                    {
                        column: 'key',
                        equals: key
                    },
                    {
                        column: '_table',
                        equals: table
                    }
                ]
            });
            return true;
        }
        else
            return false;
    }
    setTimedData(table, key, duration) {
        if (!this.hasTimeoutsTable())
            throw new DatabaseError_1.DatabaseError(`There is not timeouts table instantiated.`);
        if (duration < 1)
            throw new DatabaseError_1.DatabaseError(`The provided duration is under 1ms.`);
        const tbl = this.resolveTable(table);
        if (!tbl.columns.find(c => c.primary))
            throw new DatabaseError_1.DatabaseError(`Table '${table}' has no primary key.`);
        const left = Date.now() + duration;
        __classPrivateFieldGet(this, _Database_instances, "m", _Database_createTimeout).call(this, table, key, left);
        return this;
    }
    hasTimeoutsTable() {
        return this.tables.has('__timeouts');
    }
    /**
     * Resolves a string or object into a table.
     * @param name The name of the table.
     * @param suppress Whether to suppress the error if table does not exist.
     * @returns The table found.
     */
    resolveTable(name, suppress = false) {
        var _a;
        let table = null;
        if (name instanceof Table_1.Table)
            table = name;
        if (typeof name === "string")
            table = (_a = this.tables.get(name)) !== null && _a !== void 0 ? _a : null;
        if (!table && !suppress)
            throw new DatabaseError_1.DatabaseError(`Invalid table ${name}.`);
        return table;
    }
    /**
     * Deletes garbage from this database.
     * @returns
     */
    vacuum() {
        var _a;
        return (_a = this.db) === null || _a === void 0 ? void 0 : _a.pragma("VACUUM");
    }
    /**
     * Creates a table.
     * @param name The name for this table
     * @returns The table that was created.
     */
    createTable(name) {
        const exists = this.resolveTable(name, true);
        if (exists)
            return exists;
        const table = new Table_1.Table(name, this);
        this.tables.set(table.name, table);
        return table;
    }
    /**
     * Adds a table to this database.
     * @param table The table to add.
     * @returns The database this table was added to.
     */
    addTable(table) {
        this.tables.set(table.name, table);
        return this;
    }
    /**
     * Returns the corresponding array from given map.
     * @param map The map to convert.
     * @returns The array from the map.
     */
    arrayFromMap(map) {
        return [...this.tables.values()];
    }
    get readyTimestamp() {
        return this.readyAt ? this.readyAt.getTime() : undefined;
    }
    /**
     * Gets data from a table.
     * @param name The name of the table to get its data.
     * @returns The table data if it exists.
     */
    getTableData(name) {
        var _a;
        try {
            const data = (_a = this.db) === null || _a === void 0 ? void 0 : _a.prepare(`SELECT * FROM sqlite_master WHERE type = ? AND name = ?`).get("table", name);
            return data;
        }
        catch (error) {
            console.log(error);
            return undefined;
        }
    }
    /**
     * Gets all available tables from this database.
     * @returns An array of table data.
     */
    get databaseTables() {
        var _a;
        return (_a = this.db) === null || _a === void 0 ? void 0 : _a.prepare(`
        SELECT 
            *
        FROM 
            sqlite_master 
        WHERE 
            type ='table'
        AND 
            name NOT LIKE 'sqlite_%'`).all();
    }
    /**
     * Gets multiple rows from this database.
     * @param table The table to get rows from.
     * @param options The options to apply for this query.
     * @returns The rows found from this query.
     */
    all(table, options) {
        return this.resolveTable(table).all(options);
    }
    /**
     * Gets the position of a row in a table.
     * @param table The table to get row position from.
     * @param sortOptions Options for sorting this query.
     * @param whereOptions Options to filter this query.
     * @returns The position or positions of rows (and possibly their data if multiple).
     * @example
     * const posOf = db.get("mytable", {
     *      columns: ["bank"]
     * }, {
     *      where: {
     *          column: "id",
     *          equals: "123456"
     *      }
     * })
     *
     * console.log(posOf.position)
     */
    getRowPosition(table, sortOptions, whereOptions) {
        return this.resolveTable(table).getRowPosition(sortOptions, whereOptions);
    }
    /**
     * Get total row count in a table.
     * @param table The table to get row count from.
     * @param options The options for this query.
     * @returns An object with the count property.
     */
    getRowCount(table, options) {
        return this.resolveTable(table).getRowCount(options);
    }
    set(table, data, options) {
        if (options) {
            return this.resolveTable(table).set(data, options);
        }
        const primary = this.resolveTable(table).columns.find(c => c.primary);
        if (!primary || !primary.name)
            throw new DatabaseError_1.DatabaseError(`Could not find any primary key to perform query.`);
        //@ts-ignore
        if (!data[primary.name])
            throw new DatabaseError_1.DatabaseError(`Data does not contain ${primary.name} property.`);
        //@ts-ignore
        return this.resolveTable(table).set(data, { where: { column: primary.name, equals: data[primary.name] } });
    }
    delete(table, options) {
        if (typeof options === 'object')
            return this.resolveTable(table).delete(options);
        const key = this.resolveTable(table).columns.find(c => c.primary);
        if (!key)
            throw new DatabaseError_1.DatabaseError(`Could not find any primary key to perform query.`);
        return this.resolveTable(table).delete({ where: { column: key.name, equals: options } });
    }
    insert(table, data) {
        return this.resolveTable(table).insert(data);
    }
    get(table, options) {
        if (typeof options === 'object')
            return this.resolveTable(table).get(options);
        const key = this.resolveTable(table).columns.find(c => c.primary);
        if (!key)
            throw new DatabaseError_1.DatabaseError(`Could not find any primary key to perform query.`);
        return this.resolveTable(table).get({ where: { column: key.name, equals: options } });
    }
    addTimeoutsTable() {
        this.createTable("__timeouts")
            .addColumns([
            new __1.Column()
                .setName('endsAt')
                .setType('INTEGER'),
            new __1.Column()
                .setName('key')
                .setType('TEXT')
                .setPrimary(true),
            new __1.Column()
                .setName('_table')
                .setType('TEXT')
        ]);
        return this;
    }
    /**
     * Connects the database to its corresponding file.
     * @returns The connected database.
     */
    connect() {
        return new Promise((resolve) => {
            this.db = (0, better_sqlite3_1.default)(this.options.path);
            for (const table of this.arrayFromMap(this.tables)) {
                table.create();
            }
            if (this.hasTimeoutsTable()) {
                __classPrivateFieldGet(this, _Database_instances, "m", _Database_loadTimeouts).call(this);
            }
            this.readyAt = new Date();
            this.emit("ready");
            resolve(this);
        });
    }
    /**
     * Closes this database connection.
     * @returns
     */
    close() {
        var _a;
        (_a = this.db) === null || _a === void 0 ? void 0 : _a.close();
        this.readyAt = undefined;
        return true;
    }
}
_Database_timeouts = new WeakMap(), _Database_instances = new WeakSet(), _Database_loadTimeouts = function _Database_loadTimeouts() {
    const timeouts = this.all('__timeouts');
    for (const timeout of timeouts) {
        __classPrivateFieldGet(this, _Database_instances, "m", _Database_createTimeout).call(this, timeout._table, timeout.key, timeout.endsAt);
    }
}, _Database_createTimeout = function _Database_createTimeout(table, key, endsAt) {
    return __awaiter(this, void 0, void 0, function* () {
        const total = endsAt - Date.now();
        const id = `${table}_${key}`;
        const existing = __classPrivateFieldGet(this, _Database_timeouts, "f").get(id);
        if (existing) {
            if (existing.timeout)
                clearTimeout(existing.timeout);
            if (existing.sleep)
                clearTimeout(existing.sleep);
            __classPrivateFieldGet(this, _Database_timeouts, "f").delete(id);
        }
        this.set('__timeouts', {
            key,
            endsAt,
            _table: table
        });
        if (total > __classPrivateFieldGet(this, _Database_instances, "a", _Database_maxAllowedTimeout_get)) {
            const sleep = setTimeout(() => {
                __classPrivateFieldGet(this, _Database_instances, "m", _Database_createTimeout).call(this, table, key, endsAt);
            }, __classPrivateFieldGet(this, _Database_instances, "a", _Database_maxAllowedTimeout_get));
            __classPrivateFieldGet(this, _Database_timeouts, "f").set(id, { sleep });
            return;
        }
        else {
            const timeout = setTimeout(() => {
                const data = this.get(table, key);
                this.delete(table, key);
                if (data) {
                    this.emit('expire', table, data);
                }
                this.delete('__timeouts', {
                    where: [
                        {
                            column: '_table',
                            equals: table
                        },
                        {
                            column: 'key',
                            equals: key
                        }
                    ]
                });
                __classPrivateFieldGet(this, _Database_timeouts, "f").delete(id);
            }, total);
            __classPrivateFieldGet(this, _Database_timeouts, "f").set(id, { timeout });
            return;
        }
    });
}, _Database_maxAllowedTimeout_get = function _Database_maxAllowedTimeout_get() {
    return 2073600000;
}, _Database_sleep = function _Database_sleep(duration) {
    return new Promise((resolve, reject) => {
        if (duration >= __classPrivateFieldGet(this, _Database_instances, "a", _Database_maxAllowedTimeout_get)) {
            return __classPrivateFieldGet(this, _Database_instances, "m", _Database_sleep).call(this, __classPrivateFieldGet(this, _Database_instances, "a", _Database_maxAllowedTimeout_get)).then(resolve);
        }
        else {
            return setTimeout(resolve, __classPrivateFieldGet(this, _Database_instances, "a", _Database_maxAllowedTimeout_get));
        }
    });
};
exports.default = Database;
//# sourceMappingURL=Database.js.map